"use strict";
/**
 * @license
 * Copyright (c) 2019 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt The complete set of authors may be found
 * at http://polymer.github.io/AUTHORS.txt The complete set of contributors may
 * be found at http://polymer.github.io/CONTRIBUTORS.txt Code distributed by
 * Google as part of the polymer project is also subject to an additional IP
 * rights grant found at http://polymer.github.io/PATENTS.txt
 */
Object.defineProperty(exports, "__esModule", { value: true });
const fsExtra = require("fs-extra");
const ProgressBar = require("progress");
const ansi = require("ansi-escape-sequences");
const json_output_1 = require("./json-output");
const browser_1 = require("./browser");
const csv_1 = require("./csv");
const stats_1 = require("./stats");
const format_1 = require("./format");
const github = require("./github");
const specs_1 = require("./specs");
function combineResults(results) {
    const combined = Object.assign(Object.assign({}, results[0]), { millis: [] });
    for (const result of results) {
        combined.millis.push(...result.millis);
    }
    return combined;
}
async function automaticMode(config, servers) {
    let completeGithubCheck;
    if (config.githubCheck !== undefined) {
        completeGithubCheck = await github.createCheck(config.githubCheck);
    }
    console.log('Running benchmarks\n');
    const specs = config.benchmarks;
    const bar = new ProgressBar('[:bar] :status', {
        total: specs.length * (config.sampleSize + /** warmup */ 1),
        width: 58,
    });
    const browsers = new Map();
    for (const { browser } of specs) {
        const sig = browser_1.browserSignature(browser);
        if (browsers.has(sig)) {
            continue;
        }
        bar.tick(0, { status: `launching ${browser.name}` });
        // It's important that we execute each benchmark iteration in a new tab.
        // At least in Chrome, each tab corresponds to process which shares some
        // amount of cached V8 state which can cause significant measurement
        // effects. There might even be additional interaction effects that
        // would require an entirely new browser to remove, but experience in
        // Chrome so far shows that new tabs are neccessary and sufficient.
        const driver = await browser_1.makeDriver(browser);
        const tabs = await driver.getAllWindowHandles();
        // We'll always launch new tabs from this initial blank tab.
        const initialTabHandle = tabs[0];
        browsers.set(sig, { name: browser.name, driver, initialTabHandle });
    }
    const runSpec = async (spec) => {
        let server;
        if (spec.url.kind === 'local') {
            server = servers.get(spec);
            if (server === undefined) {
                throw new Error('Internal error: no server for spec');
            }
        }
        const url = specs_1.specUrl(spec, servers, config);
        const { driver, initialTabHandle } = browsers.get(browser_1.browserSignature(spec.browser));
        let millis;
        let bytesSent = 0;
        let userAgent = '';
        // TODO(aomarks) Make maxAttempts and timeouts configurable.
        const maxAttempts = 3;
        for (let attempt = 1;; attempt++) {
            await browser_1.openAndSwitchToNewTab(driver, spec.browser);
            await driver.get(url);
            if (spec.measurement === 'fcp') {
                millis = await browser_1.pollForFirstContentfulPaint(driver);
            }
            else if (spec.measurement === 'global') {
                millis = await browser_1.pollForGlobalResult(driver, spec.measurementExpression || 'undefined');
            }
            else { // bench.start() and bench.stop() callback
                if (server === undefined) {
                    throw new Error('Internal error: no server for spec');
                }
                millis = (await server.nextResults()).millis;
            }
            // Close the active tab (but not the whole browser, since the
            // initial blank tab is still open).
            await driver.close();
            await driver.switchTo().window(initialTabHandle);
            if (server !== undefined) {
                const session = server.endSession();
                bytesSent = session.bytesSent;
                userAgent = session.userAgent;
            }
            if (millis !== undefined || attempt >= maxAttempts) {
                break;
            }
            console.log(`\n\nFailed ${attempt}/${maxAttempts} times ` +
                `to get a ${spec.measurement} measurement ` +
                (spec.measurement === 'global' ?
                    `(from '${spec.measurementExpression}') ` :
                    '') +
                `in ${spec.browser.name} from ${url}. Retrying.`);
        }
        if (millis === undefined) {
            console.log();
            throw new Error(`\n\nFailed ${maxAttempts}/${maxAttempts} times ` +
                `to get a ${spec.measurement} measurement ` +
                (spec.measurement === 'global' ?
                    `(from '${spec.measurementExpression}') ` :
                    '') +
                `in ${spec.browser.name} from ${url}. Retrying.`);
        }
        return {
            name: spec.name,
            queryString: spec.url.kind === 'local' ? spec.url.queryString : '',
            version: spec.url.kind === 'local' && spec.url.version !== undefined ?
                spec.url.version.label :
                '',
            millis: [millis],
            bytesSent,
            browser: spec.browser,
            userAgent,
        };
    };
    // Do one throw-away run per benchmark to warm up our server (especially
    // when expensive bare module resolution is enabled), and the browser.
    for (let i = 0; i < specs.length; i++) {
        const spec = specs[i];
        bar.tick(0, {
            status: `warmup ${i + 1}/${specs.length} ${format_1.benchmarkOneLiner(spec)}`,
        });
        await runSpec(spec);
        bar.tick(1);
    }
    const specResults = new Map();
    for (const spec of specs) {
        specResults.set(spec, []);
    }
    // Always collect our minimum number of samples.
    const numRuns = specs.length * config.sampleSize;
    let run = 0;
    for (let sample = 0; sample < config.sampleSize; sample++) {
        for (const spec of specs) {
            bar.tick(0, {
                status: `${++run}/${numRuns} ${format_1.benchmarkOneLiner(spec)}`,
            });
            specResults.get(spec).push(await runSpec(spec));
            if (bar.curr === bar.total - 1) {
                // Note if we tick with 0 after we've completed, the status is
                // rendered on the next line for some reason.
                bar.tick(1, { status: 'done' });
            }
            else {
                bar.tick(1);
            }
        }
    }
    const makeResults = () => {
        const results = [];
        for (const sr of specResults.values()) {
            results.push(combineResults(sr));
        }
        const withStats = results.map((result) => ({
            result,
            stats: stats_1.summaryStats(result.millis),
        }));
        return stats_1.computeDifferences(withStats);
    };
    let hitTimeout = false;
    if (config.timeout > 0) {
        console.log();
        const timeoutMs = config.timeout * 60 * 1000; // minutes -> millis
        const startMs = Date.now();
        let run = 0;
        let sample = 0;
        let elapsed = 0;
        while (true) {
            if (stats_1.horizonsResolved(makeResults(), config.horizons)) {
                console.log();
                break;
            }
            if (elapsed >= timeoutMs) {
                hitTimeout = true;
                break;
            }
            // Run batches of 10 additional samples at a time for more presentable
            // sample sizes, and to nudge sample sizes up a little.
            for (let i = 0; i < 10; i++) {
                sample++;
                for (const spec of specs) {
                    run++;
                    elapsed = Date.now() - startMs;
                    const remainingSecs = Math.max(0, Math.round((timeoutMs - elapsed) / 1000));
                    const mins = Math.floor(remainingSecs / 60);
                    const secs = remainingSecs % 60;
                    process.stdout.write(`\r${format_1.spinner[run % format_1.spinner.length]} Auto-sample ${sample} ` +
                        `(timeout in ${mins}m${secs}s)` + ansi.erase.inLine(0));
                    specResults.get(spec).push(await runSpec(spec));
                }
            }
        }
    }
    // Close the browsers by closing each of their last remaining tabs.
    await Promise.all([...browsers.values()].map(({ driver }) => driver.close()));
    const withDifferences = makeResults();
    console.log();
    const { fixed, unfixed } = format_1.automaticResultTable(withDifferences);
    console.log(format_1.horizontalTermResultTable(fixed));
    console.log(format_1.verticalTermResultTable(unfixed));
    if (hitTimeout === true) {
        console.log(ansi.format(`[bold red]{NOTE} Hit ${config.timeout} minute auto-sample timeout` +
            ` trying to resolve horizon(s)`));
        console.log('Consider a longer --timeout or different --horizon');
    }
    if (config.jsonFile) {
        const json = await json_output_1.jsonOutput(withDifferences);
        await fsExtra.writeJSON(config.jsonFile, json, { spaces: 2 });
    }
    // TOOD(aomarks) Remove this in next major version.
    if (config.legacyJsonFile) {
        const json = await json_output_1.legacyJsonOutput(withDifferences.map((s) => s.result));
        await fsExtra.writeJSON(config.legacyJsonFile, json);
    }
    if (config.csvFile) {
        await fsExtra.writeFile(config.csvFile, csv_1.formatCsv(withDifferences));
    }
    if (completeGithubCheck !== undefined) {
        const markdown = format_1.horizontalHtmlResultTable(fixed) + '\n' +
            format_1.verticalHtmlResultTable(unfixed);
        await completeGithubCheck(markdown);
    }
    return withDifferences;
}
exports.automaticMode = automaticMode;
//# sourceMappingURL=automatic.js.map