/**
 * @license
 * Copyright (c) 2019 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt The complete set of authors may be found
 * at http://polymer.github.io/AUTHORS.txt The complete set of contributors may
 * be found at http://polymer.github.io/CONTRIBUTORS.txt Code distributed by
 * Google as part of the polymer project is also subject to an additional IP
 * rights grant found at http://polymer.github.io/PATENTS.txt
 */
import * as webdriver from 'selenium-webdriver';
/** Tachometer browser names. Often but not always equal to WebDriver's. */
export declare type BrowserName = 'chrome' | 'firefox' | 'safari' | 'edge' | 'ie';
/** Browsers we can drive. */
export declare const supportedBrowsers: Set<BrowserName>;
/** Browsers for which we can find the first contentful paint (FCP) time. */
export declare const fcpBrowsers: Set<BrowserName>;
export interface BrowserConfig {
    /** Name of the browser. */
    name: BrowserName;
    /** Whether to run in headless mode. */
    headless: boolean;
    /** A remote WebDriver server to launch the browser from. */
    remoteUrl?: string;
    /** Launch the browser window with these dimensions. */
    windowSize: WindowSize;
    /** Path to custom browser binary. */
    binary?: string;
    /** Additional binary arguments. */
    addArguments?: string[];
    /** WebDriver default binary arguments to omit. */
    removeArguments?: string[];
    /** CPU Throttling rate. (1 is no throttle, 2 is 2x slowdown, etc). */
    cpuThrottlingRate?: number;
    /** Advanced preferences usually set from the about:config page. */
    preferences?: {
        [name: string]: string | number | boolean;
    };
}
export interface WindowSize {
    width: number;
    height: number;
}
/**
 * Create a deterministic unique string key for the given BrowserConfig.
 */
export declare function browserSignature(config: BrowserConfig): string;
declare type BrowserConfigWithoutWindowSize = Pick<BrowserConfig, Exclude<keyof BrowserConfig, 'windowSize'>>;
/**
 * Parse and validate a browser string specification. Examples:
 *
 *   chrome
 *   chrome-headless
 *   chrome@<remote-selenium-server>
 */
export declare function parseBrowserConfigString(str: string): BrowserConfigWithoutWindowSize;
/**
 * Throw if any property of the given BrowserConfig is invalid.
 */
export declare function validateBrowserConfig({ name, headless, remoteUrl, windowSize }: BrowserConfig): void;
/**
 * Configure a WebDriver suitable for benchmarking the given browser.
 */
export declare function makeDriver(config: BrowserConfig): Promise<webdriver.WebDriver>;
/**
 * Open a new tab and switch to it. Assumes that the driver is on a page that
 * hasn't replaced `window.open` (e.g. the initial blank tab that we always
 * switch back to after running a benchmark).
 */
export declare function openAndSwitchToNewTab(driver: webdriver.WebDriver, config: BrowserConfig): Promise<void>;
/**
 * Return the First Contentful Paint (FCP) time (millisecond interval since
 * navigation) for the given driver. Polls every 100 milliseconds, and returns
 * undefined if no FCP was found after 10 seconds.
 *
 * https://w3c.github.io/paint-timing/#first-contentful-paint
 * https://developers.google.com/web/tools/lighthouse/audits/first-contentful-paint
 */
export declare function pollForFirstContentfulPaint(driver: webdriver.WebDriver): Promise<number | undefined>;
/**
 * Poll for the `window.tachometerResult` global and return it once it is set.
 * Polls every 50 milliseconds, and returns undefined if no result was found
 * after 10 seconds. Throws if a value was found, but it was not a number, or it
 * was a negative number.
 */
export declare function pollForGlobalResult(driver: webdriver.WebDriver, expression: string): Promise<number | undefined>;
export {};
