"use strict";
/**
 * @license
 * Copyright (c) 2019 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt The complete set of authors may be found
 * at http://polymer.github.io/AUTHORS.txt The complete set of contributors may
 * be found at http://polymer.github.io/CONTRIBUTORS.txt Code distributed by
 * Google as part of the polymer project is also subject to an additional IP
 * rights grant found at http://polymer.github.io/PATENTS.txt
 */
Object.defineProperty(exports, "__esModule", { value: true });
// It's not ideal that these WebDriver plugin packages are dependencies of the
// tachometer package, since on install they download binaries for each plugin.
// IE in particular is rarely used, so is particularly wasteful. An alternative
// might be to not depend on any of these packages, and instead prompt the user
// to install them only the first time they try to drive a browser that we
// detect there is no plugin for.
//
// Also note that the edgedriver package doesn't work on recent versions of
// Windows 10, so users must manually install following Microsoft's
// documentation.
require('chromedriver');
require('geckodriver');
require('iedriver');
const webdriver = require("selenium-webdriver");
const chrome = require("selenium-webdriver/chrome");
const firefox = require("selenium-webdriver/firefox");
const edge = require("selenium-webdriver/edge");
const util_1 = require("./util");
/** Browsers we can drive. */
exports.supportedBrowsers = new Set(['chrome', 'firefox', 'safari', 'edge', 'ie']);
/** Cases where Tachometer's browser name scheme does not equal WebDriver's. */
const webdriverBrowserNames = new Map([
    ['edge', 'MicrosoftEdge'],
    ['ie', 'internet explorer'],
]);
/** Browsers that support headless mode. */
const headlessBrowsers = new Set(['chrome', 'firefox']);
/** Browsers for which we can find the first contentful paint (FCP) time. */
exports.fcpBrowsers = new Set(['chrome']);
/**
 * Create a deterministic unique string key for the given BrowserConfig.
 */
function browserSignature(config) {
    return JSON.stringify([
        config.name,
        config.headless,
        config.remoteUrl || '',
        config.windowSize.width,
        config.windowSize.height,
    ]);
}
exports.browserSignature = browserSignature;
/**
 * Parse and validate a browser string specification. Examples:
 *
 *   chrome
 *   chrome-headless
 *   chrome@<remote-selenium-server>
 */
function parseBrowserConfigString(str) {
    let remoteUrl;
    const at = str.indexOf('@');
    if (at !== -1) {
        remoteUrl = str.substring(at + 1);
        str = str.substring(0, at);
    }
    const headless = str.endsWith('-headless');
    if (headless === true) {
        str = str.replace(/-headless$/, '');
    }
    const name = str;
    const config = { name, headless };
    if (remoteUrl !== undefined) {
        config.remoteUrl = remoteUrl;
    }
    return config;
}
exports.parseBrowserConfigString = parseBrowserConfigString;
/**
 * Throw if any property of the given BrowserConfig is invalid.
 */
function validateBrowserConfig({ name, headless, remoteUrl, windowSize }) {
    if (!exports.supportedBrowsers.has(name)) {
        throw new Error(`Browser ${name} is not supported, ` +
            `only ${[...exports.supportedBrowsers].join(', ')} are currently supported.`);
    }
    if (headless === true && !headlessBrowsers.has(name)) {
        throw new Error(`Browser ${name} does not support headless mode.`);
    }
    if (remoteUrl !== undefined && !util_1.isHttpUrl(remoteUrl)) {
        throw new Error(`Invalid browser remote URL "${remoteUrl}".`);
    }
    if (windowSize.width < 0 || windowSize.height < 0) {
        throw new Error(`Invalid window size, width and height must be >= 0.`);
    }
}
exports.validateBrowserConfig = validateBrowserConfig;
/**
 * Configure a WebDriver suitable for benchmarking the given browser.
 */
async function makeDriver(config) {
    const builder = new webdriver.Builder();
    const webdriverName = webdriverBrowserNames.get(config.name) || config.name;
    builder.forBrowser(webdriverName);
    builder.setChromeOptions(chromeOpts(config));
    builder.setFirefoxOptions(firefoxOpts(config));
    if (config.remoteUrl !== undefined) {
        builder.usingServer(config.remoteUrl);
    }
    else if (config.name === 'edge') {
        // There appears to be bug where WebDriver doesn't automatically start or
        // find an Edge service and throws "Cannot read property 'start' of null"
        // so we need to start the service ourselves.
        // See https://stackoverflow.com/questions/48577924.
        // tslint:disable-next-line:no-any TODO setEdgeService function is missing.
        builder.setEdgeService(new edge.ServiceBuilder());
    }
    const driver = await builder.build();
    if (config.name === 'safari' || config.name === 'edge' ||
        config.name === 'ie') {
        // Safari, Edge, and IE don't have flags we can use to launch with a given
        // window size, but webdriver can resize the window after we've started up.
        await driver.manage().window().setRect(config.windowSize);
    }
    return driver;
}
exports.makeDriver = makeDriver;
function chromeOpts(config) {
    const opts = new chrome.Options();
    if (config.binary) {
        opts.setChromeBinaryPath(config.binary);
    }
    if (config.headless === true) {
        opts.addArguments('--headless');
    }
    if (config.addArguments) {
        opts.addArguments(...config.addArguments);
    }
    if (config.removeArguments) {
        opts.excludeSwitches(...config.removeArguments);
    }
    const { width, height } = config.windowSize;
    opts.addArguments(`--window-size=${width},${height}`);
    return opts;
}
function firefoxOpts(config) {
    const opts = new firefox.Options();
    if (config.preferences) {
        for (const [name, value] of Object.entries(config.preferences)) {
            opts.setPreference(name, value);
        }
    }
    if (config.binary) {
        opts.setBinary(config.binary);
    }
    if (config.headless === true) {
        // tslint:disable-next-line:no-any TODO Incorrect types.
        opts.addArguments('-headless');
    }
    const { width, height } = config.windowSize;
    // tslint:disable-next-line:no-any TODO Incorrect types.
    opts.addArguments(`-width=${width}`);
    // tslint:disable-next-line:no-any TODO Incorrect types.
    opts.addArguments(`-height=${height}`);
    return opts;
}
/**
 * Open a new tab and switch to it. Assumes that the driver is on a page that
 * hasn't replaced `window.open` (e.g. the initial blank tab that we always
 * switch back to after running a benchmark).
 */
async function openAndSwitchToNewTab(driver, config) {
    // Chrome and Firefox add new tabs to the end of the handle list, but Safari
    // adds them to the beginning. Just look for the new one instead of making
    // any assumptions about this order.
    const tabsBefore = await driver.getAllWindowHandles();
    if (tabsBefore.length !== 1) {
        throw new Error(`Expected only 1 open tab, got ${tabsBefore.length}`);
    }
    // "noopener=yes" prevents the new window from being able to access the
    // first window. We set that here because in Chrome (and perhaps other
    // browsers) we see a very significant improvement in the reliability of
    // measurements, in particular it appears to eliminate interference between
    // code across runs. It is likely this flag increases process isolation in a
    // way that prevents code caching across tabs.
    await driver.executeScript('window.open("", "", "noopener=yes");');
    const tabsAfter = await driver.getAllWindowHandles();
    const newTabs = tabsAfter.filter((tab) => tab !== tabsBefore[0]);
    if (newTabs.length !== 1) {
        throw new Error(`Expected to create 1 new tab, got ${newTabs.length}`);
    }
    await driver.switchTo().window(newTabs[0]);
    if (config.name === 'ie') {
        // For IE we get a new window instead of a new tab, so we need to resize
        // every time.
        await driver.manage().window().setRect(config.windowSize);
    }
    const driverWithSendDevToolsCommand = driver;
    if (driverWithSendDevToolsCommand.sendDevToolsCommand &&
        config.cpuThrottlingRate !== undefined) {
        // Enables CPU throttling to emulate slow CPUs.
        await driverWithSendDevToolsCommand.sendDevToolsCommand('Emulation.setCPUThrottlingRate', { rate: config.cpuThrottlingRate });
    }
}
exports.openAndSwitchToNewTab = openAndSwitchToNewTab;
/**
 * Return the First Contentful Paint (FCP) time (millisecond interval since
 * navigation) for the given driver. Polls every 100 milliseconds, and returns
 * undefined if no FCP was found after 10 seconds.
 *
 * https://w3c.github.io/paint-timing/#first-contentful-paint
 * https://developers.google.com/web/tools/lighthouse/audits/first-contentful-paint
 */
async function pollForFirstContentfulPaint(driver) {
    for (let waited = 0; waited <= 10000; waited += 100) {
        await wait(100);
        const entries = await driver.executeScript('return window.performance.getEntriesByName(' +
            '"first-contentful-paint");');
        if (entries.length > 0) {
            return entries[0].startTime;
        }
    }
}
exports.pollForFirstContentfulPaint = pollForFirstContentfulPaint;
/**
 * Poll for the `window.tachometerResult` global and return it once it is set.
 * Polls every 50 milliseconds, and returns undefined if no result was found
 * after 10 seconds. Throws if a value was found, but it was not a number, or it
 * was a negative number.
 */
async function pollForGlobalResult(driver, expression) {
    // Both here and for FCP above, we could automatically tune the poll time
    // after we get our first result, so that when the script is fast we spend
    // less time waiting, and so that when the script is slow we interfere it
    // less frequently.
    for (let waited = 0; waited <= 10000; waited += 50) {
        await wait(50);
        const result = await driver.executeScript(`return (${expression});`);
        if (result !== undefined && result !== null) {
            if (typeof result !== 'number') {
                throw new Error(`'${expression}' was type ` +
                    `${typeof result}, expected number.`);
            }
            if (result < 0) {
                throw new Error(`'${expression}' was negative: ${result}`);
            }
            return result;
        }
    }
}
exports.pollForGlobalResult = pollForGlobalResult;
const wait = (ms) => new Promise((resolve) => setTimeout(resolve, ms));
//# sourceMappingURL=browser.js.map